# In buildmanifest:
# url: https://core.voluti.com.br
# "/accounting/balanceLedger/view": ["static/chunks/pages/accounting/balanceLedger/view-225856594a4838e7.js"]
# Turns to:
# https://core.voluti.com.br/_next/static/chunks/pages/accounting/balanceLedger/view-225856594a4838e7.js
# Source is:
# https://core.voluti.com.br/_next/static/chunks/pages/accounting/balanceLedger/view-225856594a4838e7.js.map

# from https://github.com/0xM1gu3l/NextJSManifestDumper